import logging

def get_default_logger(name: str, level=logging.DEBUG):
  logger = logging.getLogger(name)
  logger.setLevel(level)
  # 创建一个处理器（例如StreamHandler，用于控制台输出）
  console_handler = logging.StreamHandler()
  # 设置处理器的级别（只有达到或高于此级别的消息才会被发送到处理器）
  console_handler.setLevel(level)
  # 格式化处理
  formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
  console_handler.setFormatter(formatter)
  # 将处理器添加到Logger中
  logger.addHandler(console_handler)
  return logger



def create(workflow, chunk_id, chunk_info):
    @workflow.chunk(chunk_id=chunk_id, **chunk_info)
    def print_executor(inputs, storage):
        if len(inputs.keys()) == 1 and "input" in inputs:
            print(
                chunk_info["placeholder"] if "placeholder" in chunk_info else ">>>",
                inputs["input"]
            )
        else:
            print(
                chunk_info["placeholder"] if "placeholder" in chunk_info else ">>>",
                inputs
            )


from .utils import RequestABC, to_prompt_structure, to_instruction, to_json_desc
from openai import AsyncAzureOpenAI
from Agently.utils import RuntimeCtxNamespace
import httpx
import time

class AzureOpenAI(RequestABC):
    def __init__(self, request):
        self.request = request
        self.use_assistant = False
        self.assistant_id = None
        self.model_name = "AzureOpenAI"
        self.model_settings = RuntimeCtxNamespace(f"model.{ self.model_name }", self.request.settings)
        self.request_type = self.request.request_runtime_ctx.get("request_type", "chat")
        if self.request_type == None:
            self.request_type = "chat"

    def _create_client(self):
        client_params = {}

        proxy = self.request.settings.get_trace_back("proxy")
        if proxy:
            client_params.update({
                "http_client": httpx.AsyncClient(
                    proxies = proxy,
                    headers = [("Connection", "close")],
                )
            })
        else:
            client_params.update({
                "http_client": httpx.AsyncClient(
                    headers = [("Connection", "close")],
                )
            })

        auth = self.model_settings.get_trace_back("auth", {})
        if "api_key" not in auth or "api_version" not in auth and "azure_endpoint" not in auth:
            raise Exception("[Request: AzureOpenAI] Missing required auth items: [`api_key`, `api_version`, `azure_endpoint`]. Use .set_settings('model.AzureOpenAI.auth', { 'api_key': 'xxxxxx', 'api_version': 'xxxxxx', 'azure_endpoint': 'xxxxxx' }) to state.")
        client_params.update(auth)

        client = AsyncAzureOpenAI(**client_params)
        return client

    def construct_request_messages(self):
        #init request messages
        request_messages = []
        # - general instruction
        general_instruction_data = self.request.request_runtime_ctx.get_trace_back("prompt.general")
        if general_instruction_data:
            request_messages.append({ "role": "system", "content": f"[GENERAL INSTRUCTION]\n{ to_instruction(general_instruction_data) }" })
        # - role
        role_data = self.request.request_runtime_ctx.get_trace_back("prompt.role")
        if role_data:
            request_messages.append({ "role": "system", "content": f"[ROLE SETTINGS]\n{ to_instruction(role_data) }" })
        # - user info
        user_info_data = self.request.request_runtime_ctx.get_trace_back("prompt.user_info")
        if user_info_data:
            request_messages.append({ "role": "system", "content": f"[ABOUT USER]\n{ to_instruction(user_info_data) }" })
        # - abstract
        headline_data = self.request.request_runtime_ctx.get_trace_back("prompt.abstract")
        if headline_data:
            request_messages.append({ "role": "assistant", "content": to_instruction(headline_data) })
        # - chat history
        chat_history_data = self.request.request_runtime_ctx.get_trace_back("prompt.chat_history")
        if chat_history_data:
            request_messages.extend(chat_history_data)
        # - request message (prompt)
        prompt_input_data = self.request.request_runtime_ctx.get_trace_back("prompt.input")
        prompt_info_data = self.request.request_runtime_ctx.get_trace_back("prompt.info")
        prompt_instruct_data = self.request.request_runtime_ctx.get_trace_back("prompt.instruct")
        prompt_output_data = self.request.request_runtime_ctx.get_trace_back("prompt.output")
        # - files
        files_data = self.request.request_runtime_ctx.get_trace_back("prompt.files")
        # --- only input
        if not prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            raise Exception("[Request] Missing 'prompt.input', 'prompt.info', 'prompt.instruct', 'prompt.output' in request_runtime_ctx. At least set value to one of them.")
        prompt_text = ""
        if prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            prompt_text = to_instruction(prompt_input_data)
        # --- construct prompt
        else:
            prompt_dict = {}
            if prompt_input_data:
                prompt_dict["[INPUT]"] = to_instruction(prompt_input_data)
            if prompt_info_data:
                prompt_dict["[HELPFUL INFORMATION]"] = str(prompt_info_data)
            if prompt_instruct_data:
                prompt_dict["[INSTRUCTION]"] = to_instruction(prompt_instruct_data)
            if prompt_output_data:
                if isinstance(prompt_output_data, (dict, list, set)):
                    prompt_dict["[OUTPUT REQUIREMENT]"] = {
                        "TYPE": "JSON can be parsed in Python",
                        "FORMAT": to_json_desc(prompt_output_data),
                    }
                    self.request.request_runtime_ctx.set("response:type", "JSON")
                else:
                    prompt_dict["[OUTPUT REQUIERMENT]"] = str(prompt_output_data)
            prompt_text = to_prompt_structure(prompt_dict, end="[OUTPUT]:\n")
        # --- for vision model
        if self.request_type == "vision" and files_data:
            content = []
            content.append({ "type": "text", "text": prompt_text })
            for image_content in files_data:
                if isinstance(image_content, str):
                    content.append({ "type": "image_url", "image_url": { "url": image_content } })
                elif isinstance(image_content, dict):
                    content.append({ "type": "image_url", "image_url": image_content })
                else:
                    raise Exception(f"[Request] OpenAI vision model .files() can only accept string or dict.")
            request_messages.append({ "role": "user", "content": content })
        else:
            request_messages.append({ "role": "user", "content": prompt_text })
        return request_messages                

    def generate_request_data(self):
        self.use_assistant = self.model_settings.get_trace_back("use_assistant", False)
        if self.request_type == "chat" and self.use_assistant:
            assistant_id = self.model_settings.get_trace_back("assistant_id")
            if not assistant_id:
                raise Exception("[Request] OpenAI require 'assistant_id' when 'use_assistant' is True. Use agent.OpenAIAssistant.update() to create an assistant if you don't have one.")
            else:
                self.assistant_id = assistant_id
            options = self.model_settings.get_trace_back("options", {})
            if "model" not in options:
                options.update({ "model": "gpt-35-turbo" })
            return {
                "stream": True,
                "messages": self.construct_request_messages(),
                "options": options,
            }
        elif self.request_type == "chat":
            options = self.model_settings.get_trace_back("options", {})
            if "model" not in options:
                options.update({ "model": "gpt-35-turbo" })
            return {
                "stream": True,
                "messages": self.construct_request_messages(),
                **options
            }
        elif self.request_type == "vision":
            options = self.model_settings.get_trace_back("options", {})
            if "model" not in options:
                options.update({ "model": "gpt-4" })
            return {
                "stream": True,
                "messages": self.construct_request_messages(),
                **options
            }

    async def request_assiatant(self, request_data: dict):
        client = self._create_client()
        # create thread
        thread = await client.beta.threads.create(
            messages = request_data["messages"]
        )
        # create run
        run = await client.beta.threads.runs.create(
            thread_id = thread.id,
            assistant_id = self.assistant_id,
            **request_data["options"]
        )
        # wait until run finished
        is_complete = False
        retrieved_run = None
        while not is_complete:
            time.sleep(1)
            retrieved_run = await client.beta.threads.runs.retrieve(
                run_id = run.id,
                thread_id = thread.id,
            )
            if retrieved_run.status in ("completed", "requires_action"):
                is_complete = True
        # retrieve run step
        retrieved_run_steps = await client.beta.threads.runs.steps.list(
            run_id = run.id,
            thread_id = thread.id,
        )
        message_id = None
        # find complete or requires_action step to get message_id
        for step in retrieved_run_steps.data:
            if step.status in ("completed", "requires_action"):
                message_id = step.step_details.message_creation.message_id
                break
        # retrieve message
        message = await client.beta.threads.messages.retrieve(
            message_id = message_id,
            thread_id = thread.id,
        )
        # finally we get it...
        return message.content[0].text.value

    async def request_gpt(self, request_data: dict):
        client = self._create_client()
        #if self.request.request_runtime_ctx.get("response:type") == "JSON":
            #request_data.update({ "response_format": { "type": "json_object" } })
        stream = await client.chat.completions.create(
            **request_data
        )
        return stream

    async def request_vision(self, request_data: dict):
        client = self._create_client()
        request_data["max_tokens"] = 4096
        stream = await client.chat.completions.create(
            **request_data
        )
        return stream

    async def request_model(self, request_data: dict):
        if self.use_assistant and self.request_type == "chat":
            return await self.request_assiatant(request_data)
        elif self.request_type == "chat":
            return await self.request_gpt(request_data)
        elif self.request_type == "vision":
            return await self.request_vision(request_data)

    async def broadcast_response_without_streaming(self, response):
        yield({ "event": "response:done", "data": response })

    async def broadcast_response_with_streaming(self, response_generator):
        response_message = {}
        async for part in response_generator:
            if "choices" in dir(part) and isinstance(part.choices, list) and len(part.choices) > 0:
                delta = dict(part.choices[0].delta)
                for key, value in delta.items():
                    if key not in response_message:
                        response_message[key] = value or ""
                    else:
                        response_message[key] += value or ""
                yield({ "event": "response:delta_origin", "data": part })
                yield({ "event": "response:delta", "data": part.choices[0].delta.content or "" })
            else:
                if self.request.settings.get_trace_back("is_debug"):
                    print(f"[Request] Azure OpenAI Server Response Message: { str(dict(part)) }")
                yield({ "event": "response:delta_origin", "data": part })
        yield({ "event": "response:done_origin", "data": response_message })
        yield({ "event": "response:done", "data": response_message["content"] })

    async def broadcast_response(self, response_generator):
        if self.use_assistant and self.request_type == "chat":
            return self.broadcast_response_without_streaming(response_generator)
        elif self.request_type == "chat":
            return self.broadcast_response_with_streaming(response_generator)
        elif self.request_type == "vision":
            return self.broadcast_response_with_streaming(response_generator)

    def export(self):
        return {
            "generate_request_data": self.generate_request_data,
            "request_model": self.request_model,
            "broadcast_response": self.broadcast_response,
        }

def export():
    return ("AzureOpenAI", AzureOpenAI)


from configparser import ConfigParser

from ..utils import PluginManager, ToolManager, RuntimeCtx
from .._global import global_plugin_manager, global_storage, global_settings, global_tool_manager, global_websocket_server
from .Agent import Agent

class AgentFactory(object):
    def __init__(
            self,
            *,
            parent_plugin_manager: object=global_plugin_manager,
            parent_tool_manager: object=global_tool_manager,
            parent_settings: object=global_settings,
            is_debug: bool=False
        ):
        #runtime ctx
        self.factory_agent_runtime_ctx = RuntimeCtx()
        self.settings = RuntimeCtx(parent = parent_settings)

        #use plugin manager
        self.plugin_manager = PluginManager(parent = parent_plugin_manager)

        #use tool mananger
        self.tool_manager = ToolManager(parent = parent_tool_manager)

        #use global storage
        self.global_storage = global_storage

        #use global websocket server
        self.global_websocket_server = global_websocket_server

        #debug
        self.set_settings("is_debug", is_debug)

    def create_agent(self, agent_id: str=None, is_debug: bool=False):
        return Agent(
            agent_id = agent_id,
            parent_agent_runtime_ctx = self.factory_agent_runtime_ctx,
            parent_tool_manager = self.tool_manager,
            global_storage = self.global_storage,            
            global_websocket_server = self.global_websocket_server,
            parent_plugin_manager = self.plugin_manager,
            parent_settings = self.settings,
            is_debug = is_debug
        )

    def register_plugin(self, module_name: str, plugin_name: str, plugin: callable):
        self.plugin_manager.register(module_name, plugin_name, plugin)
        return self

    def set_settings(self, settings_key: str, settings_value: any):
        self.settings.set(settings_key, settings_value)
        return self

    def set_global_variable(self, variable_name: str, variable_value: any):
        self.settings.set(f"global_variables.{ variable_name }", variable_value)
        return self

    def toggle_component(self, component_name: str, is_enabled: bool):
        self.set_settings(f"component_toggles.{ component_name }", is_enabled)
        return self

    def set_proxy(self, proxy_setting: any):
        self.set_settings("proxy", proxy_setting)
        return self



def create_connection_symbol(chunk, handle_name):
  """生成 chunk 连接字符"""
  return f"{get_display_name(chunk)}({handle_name})"

def create_symbol(chunk):
  """生成 chunk 符号"""
  return f"{chunk['id']}({get_display_name(chunk)})"

def get_display_name(chunk):
  return fix_display_text(chunk["title"] or f'chunk-{chunk["id"]}' or 'Unknow chunk')

def fix_display_text(label: str):
  """修复文本展示，避免特殊字符"""
  if not label:
      return label
  escaped_string = label.replace('"', '&quot;')
  return f'"{escaped_string}"'



import yaml as YAML
from .utils import ComponentABC
from Agently.utils import RuntimeCtxNamespace

class YAMLLoader(ComponentABC):
    def __init__(self, agent: object):
        self.agent = agent
        self.is_debug = lambda: self.agent.settings.get_trace_back("is_debug")
        self.settings = RuntimeCtxNamespace("plugin_settings.agent_component.YAMLReader", self.agent.settings)

    def transform_to_agently_style(self, target: any, variables: dict={}):
        if isinstance(target, dict):
            result = {}
            agently_tuple_list = [None, None]
            for key, value in target.items():
                if key == "$type":
                    agently_tuple_list[0] = self.transform_to_agently_style(value, variables=variables)
                elif key == "$desc":
                    agently_tuple_list[1] = self.transform_to_agently_style(value, variables=variables)
                else:
                    result.update({ key: self.transform_to_agently_style(value, variables=variables) })
            if agently_tuple_list[0] or agently_tuple_list[1]:
                result = (agently_tuple_list[0], agently_tuple_list[1])
            return result
        if isinstance(target, list):
            result = []
            for item in target:
                result.append(self.transform_to_agently_style(item, variables=variables))
            return result
        else:
            result = str(target)
            for key, value in variables.items():
                result = result.replace("${" + str(key) + "}", str(value))
            global_variables = self.agent.settings.get("global_variables")
            if global_variables:
                for key, value in global_variables.items():
                    result = result.replace("${" + str(key) + "}", str(value))
            return result

    def load_yaml_prompt(self, *, path:str=None, yaml:str=None, use_agently_style:bool=True, variables:dict={}):
        # load yaml content
        yaml_dict = {}
        if variables:
            use_agently_style = True
        if not path and not yaml:
            raise Exception(f"[Agent Component: YAMLReader]: one parameter between `path` or `yaml` must be provided.")
        try:
            if path:
                with open(path, "r", encoding="utf-8") as yaml_file:
                    yaml_dict = YAML.safe_load(yaml_file)
                    if use_agently_style:
                        yaml_dict = self.transform_to_agently_style(yaml_dict, variables=variables)
            else:
                yaml_dict = YAML.safe_load(yaml)
                if use_agently_style:
                    yaml_dict = self.transform_to_agently_style(yaml_dict, variables=variables)
        except Exception as e:
            raise Exception(f"[Agent Component: YAMLReader]: Error occured when read YAML from path '{ path }'.\nError: { str(e) }")
        # run agent alias
        agent_alias_list = dir(self.agent)
        for alias, value in yaml_dict.items():
            if alias in agent_alias_list:
                args = []
                kwargs = {}                    
                if isinstance(value, dict) and ("$args" in value or "$kwargs" in value):
                    if "$args" in value and isinstance(value["$args"], list):
                        args = value["$args"].copy()
                    if "$kwargs" in value and isinstance(value["$kwargs"], dict):
                        kwargs = value["$kwargs"].copy()
                    getattr(self.agent, alias)(*args, **kwargs)
                else:
                    getattr(self.agent, alias)(value)
        return self.agent

    def export(self):
        return {
            "alias": {
                "load_yaml_prompt": { "func": self.load_yaml_prompt },
            }
        }

def export():
    return ("YAMLLoader", YAMLLoader)


import uuid
import copy
import inspect
from .utils.find import has_target_by_attr
from .lib.constants import DEFAULT_OUTPUT_HANDLE_VALUE, DEFAULT_INPUT_HANDLE_VALUE, EXECUTOR_TYPE_NORMAL, EXECUTOR_TYPE_START, EXECUTOR_TYPE_END, EXECUTOR_TYPE_LOOP

# 特殊的内置的 chunk 类型
SPECIAL_CHUNK_TYPES = [
    EXECUTOR_TYPE_START,
    EXECUTOR_TYPE_END,
    EXECUTOR_TYPE_LOOP
]

class SchemaChunk:
    """
    Workflow Schema 中的单个 chunk 结构，用于辅助提供操作的 API，实际运行时，提供 chunk 描述参数交给 MainExecutor 执行。
    注意，仅当connect_to/loop 调用时，才会实质创建运行时相关参数，否则仅仅是参数挂载
    """

    def __init__(self, workflow_schema = None, type = EXECUTOR_TYPE_NORMAL, executor: callable = None, **chunk_desc):
        """
        必选参数 executor，或 type 为 'Start'
        """
        # 校验必填字段(要么 type 为特殊类型，否则必须包含 executor)
        if not type or (type not in SPECIAL_CHUNK_TYPES):
            if not executor:
                raise ValueError("Missing required key: 'executor'")

        self.chunk = {
            'id': chunk_desc.get('id', str(uuid.uuid4())),
            'title': chunk_desc.get('title'),
            'type': type or EXECUTOR_TYPE_NORMAL,
            'executor': executor,
            'handles': self._fix_handles(chunk_desc.get('handles')),
            # 连接条件（默认无条件连通）
            'connect_condition': chunk_desc.get('connect_condition') or None,
            # 连接条件的描述（if/elif，condition_id 信息等）
            'connect_condition_detail': chunk_desc.get('connect_condition_detail') or None,
            # 当前激活的待连接的 handle 名
            'active_handle': chunk_desc.get('active_handle') or None,
            # 扩展信息
            'extra_info': chunk_desc.get('extra_info') or {}
        }

        self.workflow_schema = workflow_schema

        # 调用链上，分支路径
        self.branch_chunk_stack = chunk_desc.get('branch_chunk_stack') or []
    
    def handle(self, name: str) -> 'SchemaChunk':
        """Chunk 的连接点，注意 handle 可能在此时还未注册，在其进行连接操作时，会自动帮其注册"""
        # 创建影子 chunk，逻辑分支继续延续
        shadow_chunk = self.create_shadow_chunk(persist_branch_chain=True)
        shadow_chunk.set_active_handle(name)
        return shadow_chunk
    
    def if_condition(self, condition: callable = None) -> 'SchemaChunk':
        """
        按条件连接
        """
        # 创建影子 chunk
        shadow_chunk = self.create_shadow_chunk(persist_branch_chain=True)
        # 逻辑分支在原有基础上，追加当前 chunk
        shadow_chunk.branch_chunk_stack.append(shadow_chunk)
        # 设置条件参数和辅助信息
        shadow_chunk.set_connect_condition(
            condition = condition,
            condition_detail = {
                # condition id
                "id": f"cid-{str(uuid.uuid4())}",
                "type": "if"
            }
        )
        return shadow_chunk

    def else_condition(self) -> 'SchemaChunk':
        """
        按当前条件的反条件连接（特别注意，else_condition 作用的 chunk 为往前追溯的最近一个 if_condition 作用的 chunk，两是同一个，两是成套存在的）
        """
        # Step1. 从当前 chunk 的 branch_chunk_stack 从后往前找，找到最近一个 if_condition 的 'root' chunk
        branch_chunk_stack = self.branch_chunk_stack.copy()
        cursor_chunk = branch_chunk_stack.pop() if len(branch_chunk_stack) else None
        # 本次 else 条件的root chunk（即 if 的那个 chunk）
        condition_root_chunk: 'SchemaChunk' = None
        while cursor_chunk:
            cursor_chunk_params = cursor_chunk.chunk
            if cursor_chunk_params.get('connect_condition') and (cursor_chunk_params.get('connect_condition_detail') or {}).get('type') == 'if':
                condition_root_chunk = cursor_chunk
                break
            cursor_chunk = branch_chunk_stack.pop() if len(branch_chunk_stack) else None
        
        if not condition_root_chunk:
            raise ValueError(
                f"The `else_condition` method must be called after the `if_condition` method.")

        # Step2. 基于 condition_root_chunk 做进一步的连接和更新操作
        # 2.1 生成 else_fn
        current_condition = condition_root_chunk.chunk.get('connect_condition')
        if not current_condition:
            def current_condition(values, store):
                return True
        def else_condition_func(values, store):
            res = not current_condition(values, store)
            return res

        # 2.2 进行条件连接
        # 在 shadow_root_chunk 上操作，重构 branch_chain 剩下的即剔除了主 if 的路径继续保留
        shadow_root_chunk = condition_root_chunk.create_shadow_chunk(persist_branch_chain=False)
        shadow_root_chunk.branch_chunk_stack = branch_chunk_stack.copy()
        # 设置条件链接信息
        shadow_root_chunk.set_connect_condition(
            condition = else_condition_func,
            condition_detail = {
                "id": (condition_root_chunk.chunk.get('connect_condition_detail') or {}).get('id') or f"cid-{str(uuid.uuid4())}",
                "type": "elif"
            }
        )
        return shadow_root_chunk

    def connect_to(self, chunk: 'SchemaChunk') -> 'SchemaChunk':
        if isinstance(chunk, str):
            chunks = self.workflow_schema.workflow.chunks
            chunk_info_list = chunk.split(".")
            if len(chunk_info_list) < 2:
                chunk_id = chunk_info_list[0]
                if chunk_id not in chunks:
                    raise Exception(f"Can not find '{ chunk_id }' in workflow.chunks.")
                return self._connect_to(chunks[chunk_id])
            else:
                chunk_id = chunk_info_list[0]
                handle_name = chunk_info_list[1]
                if chunk_id not in chunks:
                    raise Exception(f"Can not find '{ chunk_id }' in workflow.chunks.")
                return self._connect_to(chunks[chunk_id].handle(handle_name))
        else:
            return self._connect_to(chunk)

    def _connect_to(self, chunk: 'SchemaChunk') -> 'SchemaChunk':
        """
        连接到指定个 Chunk 节点，支持传入 create_chunk 创建好的实例，也支持传入 chunk dict
        """
        # 如果传入的是一个 dict 类型，自动进行实例化再连接
        target_chunk = self.workflow_schema.create_chunk(**chunk) if isinstance(chunk, dict) else chunk
        if not isinstance(target_chunk, SchemaChunk):
            raise ValueError(f"The 'chunk' parameter to 'connect_to' is invalid.")

        # 检测 chunk 是否在当前 workflow 中有注册
        if not self.workflow_schema.get_chunk(target_chunk.chunk['id']):
            raise ValueError(f"The two chunk are not mounted under the same 'workflow'.")
        
        # 连接两 chunk
        self.workflow_schema.connect_chunk(
            self.chunk['id'],
            target_chunk.chunk['id'],
            self.chunk.get('active_handle') or DEFAULT_OUTPUT_HANDLE_VALUE,
            target_chunk.chunk.get('active_handle') or DEFAULT_INPUT_HANDLE_VALUE,
            self.chunk.get('connect_condition') or None,
            self.chunk.get('connect_condition_detail') or None,
        )
        # chain 往前递进
        target_chunk_shadow = target_chunk.create_shadow_chunk(persist_branch_chain=False)
        target_chunk_shadow.branch_chunk_stack = self.branch_chunk_stack.copy()
        target_chunk_shadow.set_active_handle()
        return target_chunk_shadow

    def loop_with(self, sub_workflow)-> 'SchemaChunk':
        """遍历逐项处理，支持传入子 workflow/处理方法 作为处理逻辑"""
        async def loop_executor(inputs, store):
            input_val = inputs.get(DEFAULT_INPUT_HANDLE_VALUE)
            all_result = []
            if isinstance(input_val, list):
                for val in input_val:
                    all_result.append(await loop_unit_core(unit_val=val, store=store))
            elif isinstance(input_val, dict):
                for key, value in input_val.items():
                    all_result.append(await loop_unit_core(unit_val={
                        "key": key,
                        "value": value
                    }, store=store))
            elif isinstance(input_val, int):
                for i in range(input_val):
                    all_result.append(await loop_unit_core(unit_val=i, store=store))
            return all_result

        async def loop_unit_core(unit_val, store):
            if inspect.iscoroutinefunction(sub_workflow):
                return await sub_workflow(unit_val, store)
            elif inspect.isfunction(sub_workflow):
                return sub_workflow(unit_val, store)
            else:
                return await sub_workflow.start_async(unit_val)

        is_function = inspect.isfunction(sub_workflow) or inspect.iscoroutinefunction(sub_workflow)
        # 这里是新的 chunk，需要走 Schema 的 create 方法挂载
        loop_chunk = self.workflow_schema.create_chunk(
            type=EXECUTOR_TYPE_LOOP,
            executor=loop_executor,
            extra_info={
                # loop 的相关信息
                "loop_info": {
                    "type": 'function' if is_function else 'workflow',
                    "detail": sub_workflow.__name__ if is_function else sub_workflow.schema.compile()
                }
            }
        )
        return self.connect_to(loop_chunk)
    
    def get_raw_schema(self):
        return copy.deepcopy(self.chunk)
    
    def set_active_handle(self, name: str = None):
        """设置激活的连接点，连接时会默认连接到该点上"""
        self.chunk['active_handle'] = name
    
    def set_connect_condition(self, condition: callable = None, condition_detail: dict = None):
        """设置连接的条件"""
        self.chunk['connect_condition'] = condition
        self.chunk['connect_condition_detail'] = condition_detail
    
    def has_handle(self, name: str, from_inputs = True) -> bool:
        """判断连接手柄是否存在"""
        search_range = self.chunk.get('handles').get('inputs' if from_inputs else 'outputs', [])
        return has_target_by_attr(search_range, 'handle', name)
    
    def create_shadow_chunk(self, persist_branch_chain=False):
        """获取到当前 chunk 的拷贝份（不干扰 chunk 自身运算逻辑），persist_branch_chain 为是否保留逻辑分支链"""
        shadow_chunk = SchemaChunk(
            workflow_schema=self.workflow_schema,
            **self.get_raw_schema()
        )
        if persist_branch_chain:
            shadow_chunk.branch_chunk_stack = self.branch_chunk_stack.copy()
        return shadow_chunk

    def _fix_handles(self, custom_handles):
        """修正用户传入的 handles 配置格式"""
        # 处理默认 handle
        handles = custom_handles or { 'inputs': [], 'outputs': [] }
        if 'inputs' not in handles:
            handles['inputs'] = []
        if 'outputs' not in handles:
            handles['outputs'] = []

        # 处理 handle 点简写的情况（如：{ "inputs": ['topic', 'type'], "outputs": [{"handle": "output"}] } 中的 inputs 的定义方式）
        def fix_handle_core(handle_list):
            if not handle_list:
                return
            if handle_list and len(handle_list):
                for idx, handle in enumerate(handle_list):
                    if isinstance(handle, str):
                        handle_list[idx] = { "handle": handle }

        fix_handle_core(handles['inputs'])
        fix_handle_core(handles['outputs'])
        return handles



from .utils import ComponentABC
from Agently.utils import RuntimeCtxNamespace

class Role(ComponentABC):
    def __init__(self, agent: object):
        self.agent = agent
        self.role_runtime_ctx = RuntimeCtxNamespace("role", self.agent.agent_runtime_ctx)
        self.role_storage = self.agent.global_storage.table("role")

    def set_id(self, role_id: str):
        self.role_runtime_ctx.set("$ID", role_id)
        return self.agent

    def set(self, key: any, value: any=None):
        if value is not None:
            self.role_runtime_ctx.set(key, value)
        else:
            self.role_runtime_ctx.set("DESC", key)
        return self.agent

    def update(self, key: any, value: any=None):
        if value is not None:
            self.role_runtime_ctx.update(key, value)
        else:
            self.role_runtime_ctx.update("DESC", key)
        return self.agent        

    def append(self, key: any, value: any=None):
        if value is not None:
            self.role_runtime_ctx.append(key, value)
        else:
            self.role_runtime_ctx.append("DESC", key)
        return self.agent

    def extend(self, key: any, value: any=None):
        if value is not None:
            self.role_runtime_ctx.extend(key, value)
        else:
            self.role_runtime_ctx.extend("DESC", key)
        return self.agent

    def save(self, role_id: str=None):
        role_id = role_id or self.role_runtime_ctx.get("$ID")
        if role_id is None or role_id == "":
            raise Exception("[Agent Component: Role] Role ID must be stated before save. Use .set_role_id() to specific it or pass role id into .save_role(<role_id>).")
        self.role_storage.set(role_id, self.role_runtime_ctx.get()).save()
        return self.agent
            
    def load(self, role_id: str):
        role_id = role_id or self.role_runtime_ctx.get("$ID")
        if role_id is None or role_id == "":
            raise Exception("[Agent Component: Role] Role ID must be stated before load. Use .set_role_id() to specific it or pass role id into .load_role(<role_id>).")
        role_data = self.role_storage.get(role_id)
        for key, value in role_data.items():
            self.role_runtime_ctx.update(key, value)
        return self.agent

    def _prefix(self):
        role_settings = self.role_runtime_ctx.get() or {}
        if "$ID" in role_settings:
            del role_settings["$ID"]
        return {
            "role": role_settings,
        }

    def export(self):
        return {
            "prefix": self._prefix,
            "suffix": None,
            "alias": {
                "set_role_id": { "func": self.set_id },
                "set_role": { "func": self.set },
                "update_role": { "func": self.update },
                "append_role": { "func": self.append },
                "extend_role": { "func": self.extend },
                "save_role": { "func": self.save },
                "load_role": { "func": self.load },
            },
        }

def export():
    return ("Role", Role)


import asyncio
import threading
from concurrent.futures import Future

def run_async(coro):
  """
  在新的事件循环中运行异步任务并返回结果。
  
  参数:
  coro: 协程对象
  
  返回:
  协程的结果
  """
  # 如果是已经在一个事件循环中，尝试另起一个线程，在该线程中建新的事件循环来执行协程（一个线程里只能有一个事件循环）
  future = Future()

  def runner(coro, future):
    try:
      loop = asyncio.new_event_loop()
      asyncio.set_event_loop(loop)
      result = loop.run_until_complete(coro)
      future.set_result(result)
    except Exception as e:
      future.set_exception(e)
    finally:
      loop.close()

  thread = threading.Thread(target=runner, args=(coro, future))
  thread.start()
  thread.join()

  return future.result()


import os
import json
from .utils import StorageABC

class FileStorage(StorageABC):
    def __init__(self, db_name: str="default"):
        self.db_name = db_name
        self.path = "./.Agently"
        if not os.path.exists(self.path):
            os.mkdir(self.path)

    def __load_from_file(self, table_name: str):
        if not os.path.exists(f"{ self.path }/{ self.db_name }/{ table_name }.data"):
            return {}
        else:
            with open(f"{ self.path }/{ self.db_name }/{ table_name }.data", "r") as file:
                try:
                    result = json.loads(file.read())
                    return result
                except Exception as e:
                    print(f"[FileStorage] Error occurred when try to read file storage '{ self.db_name }/{ table_name }.data': { str(e) }")
                    print(f"[FileStorage] '{ self.db_name }/{ table_name }.data' will be reset to {{}}.")
                    self.__save_to_file(table_name, {})
                    return {}

    def __save_to_file(self, table_name: str, value: any):        
        if not os.path.exists(f"{ self.path }/{ self.db_name }"):
            os.mkdir(f"{ self.path }/{ self.db_name }")
        if not isinstance(value, str):
            value = json.dumps(value)
        with open(f"{ self.path }/{ self.db_name }/{ table_name }.data", "w") as file:
            file.write(value)
            return True

    def set(self, table_name: str, key: str, value: any):
        table_data = self.__load_from_file(table_name)
        table_data.update({ key: json.dumps(value) })
        self.__save_to_file(table_name, table_data)

    def set_all(self, table_name:str, full_data: dict):
        for key, value in full_data.items():
            full_data[key] = json.dumps(value)
        self.__save_to_file(table_name, full_data)

    def remove(self, table_name: str, key: str):
        table_data = self.__load_from_file(table_name)
        if key in table_data:
            del table_data[key]
        self.__save_to_file(table_name, table_data)

    def update(self, table_name:str, update_data: dict):
        table_data = self.__load_from_file(table_name)
        for key, value in update_data.items():
            table_data.update({ key: json.dumps(value) })
        self.__save_to_file(table_name, table_data)

    def get(self, table_name: str, key: str):
        table_data = self.__load_from_file(table_name)
        if key in table_data:
            return json.loads(table_data[key])
        else:
            return None

    def get_all(self, table_name: str, keys: (list, None)=None):
        table_data = self.__load_from_file(table_name)
        if keys:            
            result = {}
            for key in keys:
                if key in table_data:
                    result.update({ key: json.loads(table_data[key]) })
                else:
                    result.update({ key: None })
            return result
        else:
            for key, value in table_data.items():
                table_data[key] = json.loads(value)
            return table_data

def export():
    return ("FileStorage", FileStorage)


import asyncio
import json
from datetime import datetime
from .utils import ComponentABC
from Agently.utils import RuntimeCtxNamespace, to_json_desc

class Tool(ComponentABC):
    def __init__(self, agent: object):
        self.agent = agent
        self.is_debug = lambda: self.agent.settings.get_trace_back("is_debug")
        self.settings = RuntimeCtxNamespace("plugin_settings.agent_component.Tool", self.agent.settings)
        self.tool_manager = self.agent.tool_manager
        self.tool_dict = {}
        self.must_call_tool_info = None
        self.get_tool_func = self.tool_manager.get_tool_func
        self.call_tool_func = self.tool_manager.call_tool_func
        self.set_tool_proxy = self.tool_manager.set_tool_proxy

    def register_tool(self, tool_name: str, desc: any, args: dict, func: callable, *, categories: (str, list) = None, require_proxy: bool=False):
        self.tool_manager.register(tool_name, desc, args, func, categories=categories, require_proxy=require_proxy)
        self.tool_dict.update({ tool_name: { "tool_name": tool_name, "desc": desc, "args": args } })
        return self.agent

    def stop_tools(self, tool_name_list: (str, list)):
        if isinstance(tool_name_list, str):
            tool_name_list = [tool_name_list]
        for tool_name in tool_name_list:
            if tool_name in self.tool_dict:
                del self.tool_dict[tool_name]
        return self.agent

    def add_public_tools(self, tool_name_list: (str, list)):
        if isinstance(tool_name_list, str):
            tool_name_list = [tool_name_list]
        for tool_name in tool_name_list:
            tool_info = self.tool_manager.get_tool_info(tool_name, with_args=True)
            if tool_info:
                self.tool_dict.update({ tool_name: tool_info })
        return self.agent

    def add_public_categories(self, tool_category_list: (str, list)):
        if isinstance(tool_category_list, str):
            tool_category_list = [tool_category_list]
        tool_dict = self.tool_manager.get_tool_dict(categories=tool_category_list, with_args=True)
        for key, value in tool_dict.items():
            self.tool_dict.update({ key: value })
        return self.agent

    def add_all_public_tools(self):
        all_tool_dict = self.tool_manager.get_tool_dict(with_args=True)
        for key, value in all_tool_dict.items():
            self.tool_dict.update({ key: value })
        return self.agent

    def must_call(self, tool_name: str):
        tool_info = self.tool_manager.get_tool_info(tool_name, full=True)
        if tool_info:
            self.must_call_tool_info = tool_info
        else:
            raise Exception(f"[Agent Component] Tool-must call: can not find tool named '{ tool_name }'.")

    async def call_plan_func(self, tool):
        if self.is_debug():
            print("[Agent Component] Using Tools: Start tool using judgement...")
        tool_list = []
        for tool_name, tool_info in self.tool_dict.items():
            tool_list.append(tool_info)
        try:
            result = (
                await self.agent.worker_request
                    .input({
                        "target": self.agent.request.request_runtime_ctx.get("prompt")
                    })
                    .info("current date", datetime.now().date().strftime("%Y-%B-%d"))
                    .info("tools", to_json_desc(tool_list))
                    .instruct(
                        "rule",
                        [
                            "decide whether and what tools to use for achieving {input.target}.",
                            "* if use search tool, choose ONLY ONE SEARCH TOOL THAT FIT MOST",
                            "* search keywords should use same language as {input.target}",
                        ]
                    )
                    .output({
                        "need_tool": ("Boolean", "judge if you need to use tool to reply {input.target} or not?"),
                        "input_target_language": ("String", "language of {input.target}"),
                        "tools_using": ([{
                            "purpose": ("String", "what question you want to use tool to solve?"),
                            "using_tool": (
                                {
                                    "tool_name": ("String", "{tool_name} from {tools}"),
                                    "args": ("according {args} requirement in {tools}", ),
                                },
                            ),
                        }], "output [] if {need_tool} == false"),
                    })
                    .start_async()
            )
        except Exception as e:
            return { "Tools using error": str(e) }
        if "tools_using" not in result or result["tools_using"] == []:
            return None
        tool_results = {}
        for step in result["tools_using"]:
            if "using_tool" in step and isinstance(step["using_tool"], dict) and "tool_name" in step["using_tool"]:
                if self.is_debug():
                    print("[Using Tool]: ", step["using_tool"])
                tool_info = self.tool_manager.get_tool_info(step["using_tool"]["tool_name"], full=True)
                if tool_info:
                    tool_kwrags = step["using_tool"]["args"] if "args" in step["using_tool"] and isinstance(step["using_tool"]["args"], dict) else {}
                    if tool_info["require_proxy"]:
                        proxy = self.agent.settings.get_trace_back("proxy")
                        if proxy == None:
                            proxy = self.agent.tool_manager.get_tool_proxy()
                        if proxy:
                            tool_kwrags.update({ "proxy": proxy })
                    call_result = None
                    try:
                        tool_func = self.get_tool_func(tool_info["tool_name"])
                        if asyncio.iscoroutinefunction(tool_func):
                            call_result = await tool_func(**tool_kwrags)
                        else:
                            call_result = tool_func(**tool_kwrags)
                    except Exception as e:
                        call_result = str(e)
                        if self.is_debug():
                            print("[Tool Error]: ", e)
                    if call_result:
                        info_key = str(step["purpose"])
                        info_value = call_result["for_agent"] if isinstance(call_result, dict) and "for_agent" in call_result else call_result
                        tool_results[info_key] = info_value
                        if self.is_debug():
                            print("[Result]: ", info_key, info_value)
                else:
                    if self.is_debug():
                        print(f"[Result]: Can not find tool '{ step['using_tool']['tool_name'] }'")
        if len(tool_results.keys()) > 0:
            return tool_results
        else:
            return None

    def customize_call_plan(self, call_plan_func: callable):
        self.call_plan_func = call_plan_func
        return self.agent

    async def _prefix(self):
        if self.must_call_tool_info:
            if self.is_debug():
                print(f"[Agent Component] Using Tools: Must call '{ self.must_call_tool_info['tool_name'] }'")
            self.agent.request.request_runtime_ctx.remove("prompt.instruct")
            self.agent.request.request_runtime_ctx.remove("prompt.output")
            return {
                "info": {
                    "function_must_call": {
                        "tool_name": self.must_call_tool_info["tool_name"],
                        "desc": self.must_call_tool_info["desc"],
                        "args": self.must_call_tool_info["args"],
                    },
                },
                "output": {
                    "can_call": ("Boolean", "Is all information especially from {input} above enough to generate all arguments in {function_must_call.args}?"),
                    "args": ("if {can_call} == true, generate args dict according {function_must_call.args} requirement, else leave argument value as null"),                    
                    "question": ("String", "if {can_call}==false, output question for user to collecting enough information."),
                }
            }
        elif len(self.tool_dict.keys()) > 0:
            tool_results = await self.call_plan_func(self)
            if tool_results and len(tool_results.keys()) > 0:
                return {
                    "info": tool_results,
                    "instruct": "Response according {helpful information} provided.\nProvide URL for keywords in markdown format if needed.\nIf error occured or not enough information, response you don't know honestly unless you are very sure about the answer without information support."
                }
            else:
                return None
        else:
            return None

    def export(self):
        return {
            "prefix": self._prefix,
            "suffix": None,
            "alias": {
                "register_tool": { "func": self.register_tool },
                "call_tool": { "func": self.call_tool_func },
                "set_tool_proxy": { "func": self.set_tool_proxy },
                "add_public_tools": { "func": self.add_public_tools },
                "use_public_tools": { "func": self.add_public_tools },
                "add_public_categories": { "func": self.add_public_categories },
                "use_public_categories": { "func": self.add_public_categories },
                "add_all_public_tools": { "func": self.add_all_public_tools },
                "use_all_public_tools": { "func": self.add_all_public_tools },
                "must_call": { "func": self.must_call },
                "customize_call_plan": { "func": self.customize_call_plan },
            }
        }

def export():
    return ("Tool", Tool)


import asyncio
import logging
from .MainExecutor import MainExecutor
from .Schema import Schema
from .executors.install import mount_built_in_executors
from .lib.ChunkExecutorManager import ChunkExecutorManager
from .lib.constants import EXECUTOR_TYPE_NORMAL
from .lib.painter import draw_with_mermaid
from .yamlflow.yamlflow import start_yaml_from_str, start_yaml_from_path
from .utils.exec_tree import generate_executed_schema
from .utils.logger import get_default_logger
from .utils.runner import run_async
from ..utils import RuntimeCtx
from .._global import global_settings
from Agently.utils import IdGenerator

class Workflow:
    def __init__(self, *, schema_data: dict = None, settings: dict = {}, workflow_id:str=None):
        """
        Workflow，初始参数 schema_data 形如 { 'chunks': [], 'edges': [] }，handler 为要处理响应的函数
        """
        self.workflow_id = workflow_id or IdGenerator("workflow").create()
        # 处理设置
        self.settings = RuntimeCtx(parent = global_settings)
        if settings:
            self.settings.update_by_dict(settings)
        # logger
        workflow_default_logger = get_default_logger(self.workflow_id, level=logging.DEBUG if self.settings.get_trace_back("is_debug") else logging.WARN)
        self.logger = self.settings.get('logger', workflow_default_logger)
        # 初始 schema
        self.schema = Schema(
            schema_data=schema_data or {'chunks': [], 'edges': []},
            workflow=self,
            logger=self.logger
        )
        # 初始化执行器
        self.executor = MainExecutor(
            workflow_id=self.workflow_id,
            settings=self.settings,
            logger=self.logger
        )
        # Get Storage From Executor
        self.storage = self.executor.store
        # 装载内置类型
        mount_built_in_executors(self.executor)
        # Chunk Storage
        self.chunks = {}
        # Executor Manager
        self.executor_manager = ChunkExecutorManager()
        # Short Cut
        self.chunk("start", type = "Start")(lambda:None)
        self.chunk("end", type = "End")(lambda:None)
        self.connect_to = self.chunks["start"].connect_to

    def chunk(self, chunk_id: str=None, type=EXECUTOR_TYPE_NORMAL, **chunk_desc):
        is_class = chunk_desc.get("is_class", False)
        if not is_class:
            def create_chunk_decorator(func: callable):
                nonlocal chunk_id, type, chunk_desc
                if not chunk_id or not isinstance(chunk_id, str):
                    chunk_id = func.__name__
                if "title" not in chunk_desc or chunk_desc["title"] == "":
                    chunk_desc.update({ "title": chunk_id })
                return self.chunks.update({
                    chunk_id: self.schema.create_chunk(
                            executor = func,
                            type = type,
                            **chunk_desc
                        )
                })
            return create_chunk_decorator
        else:
            return self.chunk_class(chunk_id, type, **chunk_desc)

    def chunk_class(self, chunk_id: str=None, type=EXECUTOR_TYPE_NORMAL, **chunk_desc):
        def create_chunk_decorator(func: callable):
            nonlocal chunk_id, type, chunk_desc
            if not chunk_id or not isinstance(chunk_id, str):
                chunk_id = f"@{ func.__name__ }"
            if "title" not in chunk_desc or chunk_desc["title"] == "":
                chunk_desc.update({ "title": chunk_id })
            return self.chunks.update({
                chunk_id: {
                    "executor": func,
                    "type": type,
                    **chunk_desc
                }
            })
        return create_chunk_decorator

    def register_executor_func(self, executor_id: str, executor_func: callable):
        self.executor_manager.register(executor_id, executor_func)
        return self

    def executor_func(self, executor_id: str):
        def register_executor_decorator(func: callable):
            return self.executor_manager.register(executor_id, func)
        return register_executor_decorator

    def start_yaml(self, yaml_str=None, *, path=None, draw=False):
        if yaml_str:
            return start_yaml_from_str(self, yaml_str, draw=draw)
        elif path:
            return start_yaml_from_path(self, path, draw=draw)
        else:
            raise Exception("[Workflow] At least one parameter in `yaml_str` and `path` is required when using workflow.load_yaml().")

    async def start_async(self, start_data=None, *, storage=None):
        executed_schema = generate_executed_schema(self.schema.compile())
        res = await self.executor.start(executed_schema, start_data, storage=storage)
        return res

    def start(self, start_data = None, *, storage=None):
        return run_async(self.start_async(start_data, storage=storage))

    def reset_runtime_status(self):
        """重置运行数据"""
        self.executor.reset_all_runtime_status()
        return self

    def reset(self, schema_data: dict = None):
        """彻底重置（包含注册的chunk和连接关系）"""
        self.executor.reset_all_runtime_status()
        self.schema = Schema(schema_data or {'chunks': [], 'edges': []})
        return self
    
    def reset_connection(self):
        """重置链接关系(保留 chunk 注册)"""
        self.executor.reset_all_runtime_status()
        self.schema.remove_all_connection()
        return self
    
    def get_runtime_store(self):
        return self.executor.store

    def draw(self, type='mermaid'):
        """绘制出图形，默认使用 mermaid，可点击 https://mermaid-js.github.io/mermaid-live-editor/edit 粘贴查看效果"""
        return draw_with_mermaid(self.schema.compile())



import asyncio
import threading
import json
import websockets
import tornado.ioloop
import tornado.web
import tornado.websocket
import aiohttp
import time
import queue
import uuid
import copy

class WebSocketKeepAlive(object):
    def __init__(self, uri: str, client_id: str):
        self.uri = uri
        self.client_id = client_id
        self.stop_signal = threading.Event()
        self.send_signal = threading.Event()
        self.send_queue = queue.Queue()
        self.event_handlers = {}
        self.thread = None
        self.client = None

        async def send_message(websocket):
            while True:
                while not self.send_signal.is_set():
                    await asyncio.sleep(0)
                try:
                    message = self.send_queue.get_nowait()
                    await websocket.send_str(json.dumps(message))
                except queue.Empty:
                    self.send_signal.clear()
                
        async def receive_message(websocket):
            while True:
                await asyncio.sleep(0)
                if len(self.event_handlers.keys()) > 0:
                    async for message in websocket:
                        message = json.loads(message.data)
                        if message["client_id"] == self.client_id\
                        and message["client_id"] in self.event_handlers\
                        and message["event"] in self.event_handlers[message["client_id"]]:
                            for handler in self.event_handlers[message["client_id"]][message["event"]]:
                                if asyncio.iscoroutinefunction(handler):
                                    await handler(message["data"])
                                else:
                                    handler(message["data"])

        async def stop_keep_alive():
            while not self.stop_signal.is_set():
                await asyncio.sleep(0)
            asyncio.get_event_loop().stop()
            self.client = None

        async def connect_keep_alive():
            async with aiohttp.ClientSession() as session:
                async with session.ws_connect(self.uri) as websocket:
                    self.client = websocket
                    await asyncio.gather(send_message(websocket), receive_message(websocket))

        def start_event_loop():
            asyncio.set_event_loop(asyncio.new_event_loop())
            self.queue = asyncio.Queue()
            asyncio.get_event_loop().run_until_complete(connect_keep_alive())
            asyncio.get_event_loop().run_forever()       

        self.thread = threading.Thread(target=start_event_loop)
        self.thread.start()

        while not self.client:
            time.sleep(0.1)

    def send(self, event: str, data: any=None,):
        request_id = str(uuid.uuid4())
        self.send_queue.put_nowait({ "request_id": request_id, "event": event, "data": data })
        self.send_signal.set()
        return self

    def on(self, event: str, handler: callable):
        if self.client_id not in self.event_handlers:
            self.event_handlers.update({ self.client_id: {} })
        if event not in self.event_handlers[self.client_id]:
            self.event_handlers[self.client_id].update({ event: [handler] })
        else:
            self.event_handlers[self.client_id][event].append(handler)
        return self

    def stop(self):
        self.stop_signal.set()
        self.thread.join()
        self.client = None
        self.queue = None
        self.thread = None
        return self

class WebSocketClient(object):
    def __init__(self, *, host: str="0.0.0.0", port: int=15365, path: str=""):
        self.host = host
        self.port = port
        self.path = path
        self.client_id = str(uuid.uuid4())

    def create_keep_alive(self, path: str=None):
        path = path if path else self.path
        uri = f"ws://{ self.host }:{ self.port }{ path if path[0] == '/' else '/' + path }?client_id={ self.client_id }"
        return WebSocketKeepAlive(uri, self.client_id)

    async def send_async(self, event: str, data: any=None, handler: callable=None, *, path: str=None):
        path = path if path else self.path
        request_id = str(uuid.uuid4())
        uri = f"ws://{ self.host }:{ self.port }{ path if path[0] == '/' else '/' + path }?client_id={ self.client_id }"
        async with aiohttp.ClientSession() as session:
            async with session.ws_connect(uri) as websocket:
                send_message = json.dumps({ "request_id": request_id, "event": event, "data": data })
                await websocket.send_str(send_message)
                if handler != None:
                    try:
                        async for response_message in websocket:
                            response_message = json.loads(response_message.data)
                            if response_message["client_id"] == self.client_id and response_message["request_id"] == request_id:
                                if asyncio.iscoroutinefunction(handler):
                                    await handler(response_message)
                                else:
                                    handler(response_message)
                    except:
                        pass

    def send(self, event: str, data: any=None, handler: callable=None):
        asyncio.get_event_loop().run_until_complete(self.send_async(event, data, handler))
        return self

class WebSocketServer(object):
    def __init__(self, *, host: str="0.0.0.0", port: int=15365):
        self.host = host
        self.port = port
        self.stop_signal = threading.Event()
        self.is_handling = threading.Event()
        self.event_handlers = {}
        self.thread = None
        self.server = None
        self.loop = None
        self.add_handler_queue = queue.Queue()
        self.status = 0

    def set_port(self, port: int):
        self.port = port
        return self

    def start(self, port: int=None):
        def start_loop():
            self.port = port if port else self.port
            self.server = tornado.web.Application([])
            self.server.listen(port)
            if not self.add_handler_queue.empty():
                while True:
                    try:
                        self.add_handler_queue.get_nowait()()
                    except queue.Empty:
                        break
            self.loop = tornado.ioloop.IOLoop.current()
            self.status = 1
            self.loop.start()

        self.thread = threading.Thread(target=start_loop)
        self.thread.start()

        while self.status != 1:
            time.sleep(0.1)

    def stop(self):
        self.loop.stop()
        self.thread.join()
        self.status = 0
        return self

    class WebSocketRequestHandler(tornado.websocket.WebSocketHandler):
        def initialize(self, event_handlers):
            self.event_handlers = event_handlers

        def _response_message(self, event: str, data: any=None, *, close_after_response: bool=False):
            self.write_message(json.dumps({
                "client_id": self.client_id,
                "request_id": self.request_id,
                "event": event,
                "data": data,
            }))
            if close_after_response:
                self.close()
            return self.close

        async def on_message(self, message: str):
            message = json.loads(message)
            self.client_id = self.get_argument("client_id")
            self.request_id = message["request_id"]
            if message["event"] in self.event_handlers:
                for handler in self.event_handlers[message["event"]]:
                    if asyncio.iscoroutinefunction(handler):
                        await handler(message["data"], self._response_message)
                    else:
                        handler(message["data"], self._response_message)

    def add_event_handler(self, path: str, event: str, handler: callable):
        if path == None or path == "" or path == "/":
            raise Exception("[WebSocket Server] Event handler require param 'path' that is not '','/' or None.")
        if path[0] != "/":
            path = "/" + path
        if path in self.event_handlers:
            if event in self.event_handlers[path]:
                self.event_handlers[path][event].append(handler)
            else:
                self.event_handlers[path].update({ event: [handler] })
        else:
            self.event_handlers.update({ path: { event: [handler] } })
            if self.server:
                while True:
                    try:
                        self.add_handler_queue.get_nowait()()
                    except queue.Empty:
                        break
                self.server.add_handlers(".*$", [(path, self.WebSocketRequestHandler, { "event_handlers": self.event_handlers[path] })])
            else:
                self.add_handler_queue.put_nowait(lambda: self.server.add_handlers(".*$", [(path, self.WebSocketRequestHandler, { "event_handlers": self.event_handlers[path] })]))
        return self

    def remove_event_handler(self, path:str, event: str=None):
        if event == None:
            if path in self.event_handlers:
                del self.event_handlers[path]
        else:
            if path in self.event_handlers and event in self.event_handlers[path]:
                del self.event_handlers[path][event]
        self.server.remove_handlers(path)

    def __exit__(self):
        self.stop()